////////////////////////////////////////////////////////////////////////////////
// awe_sequencer.cc - AWEWavePlayer version 0.3 by Jeroen Witteman
////////////////////////////////////////////////////////////////////////////////
//
// Written by:    Jeroen Witteman
// Email adress:  jwittema@wi.leidenuniv.nl
// WWW Homepage:  http://www.wi.leidenuniv.nl/home/jwittema/
//
// copyright (c) 1998 by Jeroen Witteman, all rights reserved.
// THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!
////////////////////////////////////////////////////////////////////////////////

// includes
#include <stdlib.h>    // used for perror()
#include <unistd.h>    // used for write()
#include <linux/soundcard.h>
#include <linux/awe_voice.h>
#include "awe_sequencer.h"
#include "common.h"
#include "seq.h"
#include "version.h"
#include "config.h"

// external variables
SEQ_USE_EXTBUF();
extern int seqfd;
extern int awe_dev;
extern int max_synth_voices;
extern int nrsynths;

extern Settings settings;
extern int StereoSampleSize;
extern int NrOfSamples;

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::Initialize
// Initialize the sequencer
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::Initialize() {
  seq_init();
  SEQ_DUMPBUF();
} // AWE_Sequencer::Initialize

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::Shutdown
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::Shutdown() {
  AWE_RELEASE_ALL(awe_dev); // release all notes
  SEQ_DUMPBUF();
  seq_end(); // close the sequencer
} // AWE_Sequencer::Shutdown

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::PlayNote
// Plays a note with instrumentnr 'program' on bank 'bank' using voice channel
// 'voice'.
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::PlayNote(int voice, int bank, int program, int note) {
  AWE_SET_CHANNEL_MODE(awe_dev, AWE_PLAY_INDIRECT);
  SEQ_CONTROL(awe_dev, voice, CTL_BANK_SELECT, bank);
  SEQ_SET_PATCH(awe_dev, voice, program);

  SEQ_START_NOTE(awe_dev, voice, note, 127);

  SEQ_DUMPBUF();
} // AWE_Sequencer::PlayNote

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::StartPlaying
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::StartPlaying() {
  PlayNote(0, 0, 0, ROOT_KEY);
} // AWE_Sequencer::StartPlaying

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::ReleaseAllNotes
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::ReleaseAllNotes() {
  AWE_RELEASE_ALL(awe_dev);
  SEQ_DUMPBUF();
} // AWE_Sequencer::ReleaseAllNotes

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::ResetSamples
// Resets all the samples on your AWE (clears the memory), if reset is true in
// settings.
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::ResetSamples() {
  if (settings.reset)
    seq_reset_samples();
} // AWE_Sequencer::ResetSamples

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::OpenPatch
// Sends the open patch to your AWE (the 'awe_patch_info' header with type set
// to AWE_OPEN_PATCH and the 'awe_open_parm')
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::OpenPatch() {
  char* data;
  int length = AWE_PATCH_INFO_SIZE + AWE_OPEN_PARM_SIZE;
  awe_patch_info* p;
  awe_open_parm*  o;

  SEQ_DUMPBUF();
  data = new char[length]; // allocate memory

  p = (awe_patch_info*) data;
  o = (awe_open_parm*) (data + AWE_PATCH_INFO_SIZE);

  // set the parameters
  p->key       = AWE_PATCH;
  p->device_no = awe_dev;
  p->sf_id     = 0;  // should be zero
  p->optarg    = 0;  // not used
  p->len       = AWE_OPEN_PARM_SIZE;
  p->type      = AWE_OPEN_PATCH;
  p->reserved  = 0;

  o->type     = AWE_PAT_TYPE_MISC;   // sample type
  o->reserved = 0;
  o->name     = "AWEWavePlayer Sample           ";

  // write the data to sequencer
  if (write(seqfd, (char*)(data), length)==-1) perror("Write patch: /dev/sequencer");

  delete[] data; // deallocate the memory
} // AWE_Sequencer::OpenPatch

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::ClosePatch
// Sends the close patch to your AWE (the 'awe_patch_info' header with type set
// to AWE_CLOSE_PATCH.
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::ClosePatch() {
  char* data;
  int length = AWE_PATCH_INFO_SIZE;
  awe_patch_info* p;

  SEQ_DUMPBUF();
  data = new char[length]; // allocate memory

  p = (awe_patch_info*)data;

  p->key       = AWE_PATCH;
  p->device_no = awe_dev;
  p->sf_id     = 0;  // should be zero
  p->optarg    = 0;  // not used
  p->len       = 0;
  p->type      = AWE_CLOSE_PATCH;
  p->reserved  = 0;

  // write the data to sequencer
  if (write(seqfd, (char*)(data), length)==-1) perror("Write patch: /dev/sequencer");

  delete[] data; // deallocate memory
} // AWE_Sequencer::ClosePatch


////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::UnloadSample
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::UnloadSample() {
  char* data;
  int length = AWE_PATCH_INFO_SIZE;
  awe_patch_info* p;

  SEQ_DUMPBUF();
  data = new char[length]; // allocate memory

  p = (awe_patch_info*) data;
  p->key       = AWE_PATCH;
  p->device_no = awe_dev;
  p->sf_id     = 0;  // should be zero
  p->optarg    = 0;  // not used
  p->len       = 0;
  p->type      = AWE_UNLOAD_PATCH;
  p->reserved  = 0;

  // write the data to sequencer
  if (write(seqfd, (char*)(data), length)==-1) perror("Write patch: /dev/sequencer");

  delete[] data; // deallocate memory
} // AWE_Sequencer::UnloadSample

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::LoadSample
// Loads a sample into the AWE's memory. 'Data' is a pointer to the sample wich
// should have an empty header of 48 bytes, which are used as an info header. So
// you'll have leave the first 48 bytes blank. 'SampleSize' is the size of the
// sample in bytes without the 48 bytes header. 'sample_id' is the id used for
// this sample (used for when replacing this sample and linking the voice info
// with this sample). If 'replace' is 'true' the previous sample with
// 'sample_id' will be replaced with the new sample data.
////////////////////////////////////////////////////////////////////////////////
int AWE_Sequencer::LoadSample(void* Data, int sample_id, bool replace) {
  int Header_Size = AWE_PATCH_INFO_SIZE + AWE_SAMPLE_INFO_SIZE; // 48 bytes
  int Result;
  awe_patch_info *p;
  awe_sample_info *s;

  SEQ_DUMPBUF();

  p = (awe_patch_info*) Data;
  s = (awe_sample_info*) (Data + AWE_PATCH_INFO_SIZE);

  p->key       = AWE_PATCH;
  p->device_no = awe_dev;
  p->sf_id     = 0; // normally zero
  p->optarg    = (replace ? 28 : 0); // number of channels to use with AWE_REPLACE_DATA
                 // there are 30 channels available and we use 2 for playing (stereo)
  p->len       = AWE_SAMPLE_INFO_SIZE + (MonoSampleSize);
  p->type      = (replace ? AWE_REPLACE_DATA : AWE_LOAD_DATA);
  p->reserved  = 0;

  s->sf_id         = 0; // should be zero
  s->sample        = sample_id;
  s->start         = 0;
  s->end           = (MonoSampleSize)/2; // size in 16 bits instead of bytes (8 bits)
  s->loopstart     = 0;
  s->loopend       = (MonoSampleSize)/2;
  s->size          = (MonoSampleSize)/2;
  s->checksum_flag = 0;
  s->mode_flags    = AWE_SAMPLE_NO_BLANK;
  s->checksum      = 0;

  Result = write(seqfd, Data, (MonoSampleSize) + Header_Size);

  if (Result!=0) perror("Error during Load_Sample");

  return Result;
} // AWE_Sequencer::LoadSample

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::LoadVoiceInfo
// Loads the voice information for a sample with 'sample_id'. Gives the sample
// banknumber 'banknr' and instrumentnumber 'instrumentnr' and sets the 'pan',
// the 'chorus' and 'reverb' values.
////////////////////////////////////////////////////////////////////////////////
int AWE_Sequencer::LoadVoiceInfo(int sample_id, int banknr, int instrumentnr, int pan) {
  awe_voice_rec_patch* p;
  int nrvoice = 1;
  int Result;

  SEQ_DUMPBUF();
  p = new awe_voice_rec_patch;  // allocate memory

  p->patch.key       = AWE_PATCH;
  p->patch.device_no = awe_dev;
  p->patch.sf_id     = 0;  // normally zero
  p->patch.optarg    = 0;  // not used
  p->patch.len       = AWE_VOICE_REC_SIZE + (nrvoice * AWE_VOICE_INFO_SIZE);
  p->patch.type      = AWE_LOAD_INFO;
  p->patch.reserved  = 0;

  p->hdr.bank       = banknr;        // midi bank number
  p->hdr.instr      = instrumentnr;  // midi instrument preset number
  p->hdr.nvoices    = nrvoice;       // number of voices (samples)
  p->hdr.write_mode = AWE_WR_APPEND; // append anyway

  p->info.sf_id          = 0;  // normally zero
  p->info.sample         = sample_id;  // sample_id
  p->info.start          = 0;  // |
  p->info.end            = 0;  // | offset
  p->info.loopstart      = 0;  // | corrections
  p->info.loopend        = MonoSampleSize/2 * (NrOfSamples-1);  // add this to MonoSampleSize/2
                         // Adjust the loop to the total of all samples (it's a trick!)
  p->info.rate_offset    = 0;  // |
  p->info.mode           = 0;  // AWE_MODE_STEREO is ignored ?
  p->info.root           = ROOT_KEY;  // midi root key
  p->info.tune           = 0;   // pitch tuning
  p->info.low            = 0;   // key note range
  p->info.high           = 127; //   
  p->info.vellow         = 0;   // velocity range   
  p->info.velhigh        = 127; // 
  p->info.fixkey         = -1;
  p->info.fixvel         = 127;
  p->info.pan            = pan;  // panning
  p->info.fixpan         = pan;
  p->info.exclusiveClass = 1; 
  p->info.amplitude      = settings.volume;
  p->info.attenuation    = 0;
  p->info.scaleTuning    = 100; // normally 100

  // if p->info.mode is set at AWE_MODE_INIT_PARM these are set automatically
  // however I wish to change the chorus and reverb
  p->info.parm.moddelay    = 0x8000;  // modulation delay (0x8000)
  p->info.parm.modatkhld   = 0x7f7f;  // modulation attack & hold time (0x7f7f)
  p->info.parm.moddcysus   = 0x7f7f;  // modulation decay & sustain (0x7f7f)
  p->info.parm.modrelease  = 0x897f;  // modulation release time (0x807f)
  p->info.parm.modkeyhold  = 0;       // envelope change per key (not used)
  p->info.parm.modkeydecay = 0;        
  p->info.parm.voldelay    = 0x8000;  // volume delay (0x8000)
  p->info.parm.volatkhld   = 0x7f7f;  // volume attack & hold time (0x7f7f)
  p->info.parm.voldcysus   = 0x7f7f;  // volume decay & sustain (0x7f7f)
  p->info.parm.volrelease  = 0x807f;  // volume release time (0x807f)
  p->info.parm.volkeyhold  = 0;       // envelope change per key (not used)
  p->info.parm.volkeydecay = 0;       
  p->info.parm.lfo1delay   = 0x8000;  // LFO1 delay (0x8000)
  p->info.parm.lfo2delay   = 0x8000;  // LFO2 delay (0x8000)
  p->info.parm.pefe        = 0x0000;  // modulation pitch & cutoff (0x0000)
  p->info.parm.fmmod       = 0x0000;  // LFO1 pitch & cutoff (0x0000)
  p->info.parm.tremfrq     = 0x0000;  // LFO1 volume & freq (0x0000)
  p->info.parm.fm2frq2     = 0x0000;  // LFO2 pitch & freq (0x0000)
  p->info.parm.cutoff      = 0xff;    // initial cutoff (0xff)
  p->info.parm.filterQ     = 0x0;     // initial filter Q [0-15] (0x0)
  p->info.parm.chorus      = settings.chorus;  // chorus send (0x00)
  p->info.parm.reverb      = settings.reverb;  // reverb send (0x00)

  // write to sequencer
  Result=write(seqfd, p, sizeof(*p));

  if (Result!=0) perror("Error during Load_Voice_Info");

  delete p;  // deallocate memory

  return Result;
} // AWE_Sequencer::LoadVoiceInfo

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::DebugMode
// Sets the debug mode; value from 0 to 9
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::DebugMode(int level) {
  AWE_DEBUG_MODE(awe_dev, level);
  SEQ_DUMPBUF();
} // AWE_Sequencer::DebugMode

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::Equalizer
// Sets bass and treble control; values are from 0 to 11 
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::Equalizer(int bass, int treble) {
  AWE_EQUALIZER(awe_dev, bass, treble);
  SEQ_DUMPBUF();
} // AWE_Sequencer::Equalizer

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::ReverbMode
// Sets the reverb mode; value from 0 to 7
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::ReverbMode(int mode) {
  AWE_REVERB_MODE(awe_dev, mode);
  SEQ_DUMPBUF();
} // AWE_Sequencer::ReverbMode

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::ChorusMode
// Sets the chorus mode; value from 0 to 7
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::ChorusMode(int mode) {
  AWE_CHORUS_MODE(awe_dev, mode);
  SEQ_DUMPBUF();
} // AWE_Sequencer::ChorusMode

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::InitialAttenuation
// Sets the initial attenuation
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::InitialAttenuation(int value) {
  AWE_INITIAL_ATTEN(awe_dev, value);
  SEQ_DUMPBUF();
} // AWE_Sequencer::InitialAttenuation

////////////////////////////////////////////////////////////////////////////////
// AWE_Sequencer::MainVolume
////////////////////////////////////////////////////////////////////////////////
void AWE_Sequencer::MainVolume(int volume) {
  SEQ_CONTROL(awe_dev,    0, CTRL_MAIN_VOLUME, volume);
  SEQ_CONTROL(awe_dev, IDOS, CTRL_MAIN_VOLUME, volume);
  SEQ_DUMPBUF();
} // AWE_Sequencer::MainVolume

////////////////////////////////////////////////////////////////////////////////
// AvailableMem
////////////////////////////////////////////////////////////////////////////////
int AWE_Sequencer::AvailableMem() {
  return seq_mem_avail();
} // AvailableMem

///////////////////////////////////// eof //////////////////////////////////////
