////////////////////////////////////////////////////////////////////////////////
// interface.cc - AWEWavePlayer version 0.3 by Jeroen Witteman
////////////////////////////////////////////////////////////////////////////////
//
// Written by:    Jeroen Witteman
// Email adress:  jwittema@wi.leidenuniv.nl
// WWW Homepage:  http://www.wi.leidenuniv.nl/home/jwittema/
//
// copyright (c) 1998 by Jeroen Witteman, all rights reserved.
// THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!
////////////////////////////////////////////////////////////////////////////////

#include <iostream.h>  
#include <iomanip.h>  // used for setw()
#include <unistd.h>   // used for usleep()
#ifdef INT_NCURSES
#include <ncurses.h>
#endif
#include "interface.h"
#include "awe_sequencer.h"
#include "version.h"
#include "common.h"
#include "config.h"
#include "seq.h"

// external variables
extern AWE_Sequencer* AWE;
extern int StereoSampleSize;  // sample size in bytes
extern int NrOfSamples;      // number of samples
extern Settings settings;

////////////////////////////////////////////////////////////////////////////////
// Interface::Interface
////////////////////////////////////////////////////////////////////////////////
Interface::Interface() {
  interfacetype = Normal;
  verbosity = normal;
} // Interface::Interface

////////////////////////////////////////////////////////////////////////////////
// Interface::Initialize_interface
////////////////////////////////////////////////////////////////////////////////
void Interface::Initialize_interface() {
  FILE* keyboard;
  keyboard = fopen("/dev/tty", "r");

  switch (interfacetype) {
#ifdef INT_NCURSES
    case NCurses:
      newterm(NULL, stdout, keyboard);
      start_color();      // start color mode
      cbreak();
      noecho();
      nonl();
      curs_set(0);        // make cursor invisible
      nodelay(stdscr, 1);
      idlok(stdscr, 1);
      leaveok(stdscr, true);  // leave cursor, prevents messy display
      keypad(stdscr, true);

      init_pair(1, COLOR_BLACK,  COLOR_BLACK);
      init_pair(2, COLOR_RED,    COLOR_BLACK);
      init_pair(3, COLOR_GREEN,  COLOR_BLACK);
      init_pair(4, COLOR_YELLOW, COLOR_BLACK);
      init_pair(5, COLOR_BLUE,   COLOR_BLACK);
      init_pair(6, COLOR_MAGENTA,COLOR_BLACK);
      init_pair(7, COLOR_CYAN,   COLOR_BLACK);
      init_pair(8, COLOR_WHITE,  COLOR_BLACK);

      erase();
      refresh();
      break;
#endif
    case Normal:
      break;
  } // switch
} // Interface::Initialize_interface

////////////////////////////////////////////////////////////////////////////////
// Interface::Interactive_control
////////////////////////////////////////////////////////////////////////////////
void Interface::Interactive_control() {
#define Cur_Volume  settings.volume
#define Cur_Bass    settings.bass
#define Cur_Treble  settings.treble
  int c;

  while (true) {
    usleep(50);
    // get keyboard input
    c = wgetch(stdscr);
//     mvprintw(22, 0, "   %i   ", c);
    switch(c) {
      case 'a':  // raise volume
      case '+':
      case KEY_UP:
        Cur_Volume += 1;
        if (Cur_Volume > MAX_VOLUME) Cur_Volume = MAX_VOLUME;
        AWE->MainVolume(Cur_Volume);
        Update_volume();
        break;
      case 'z':  // lower volume
      case '-':
      case KEY_DOWN:
        Cur_Volume -= 1;
        if (Cur_Volume < MIN_VOLUME) Cur_Volume = MIN_VOLUME;
        AWE->MainVolume(Cur_Volume);
        Update_volume();
        break;

      case 'T': // raise treble
        Cur_Treble += 1;
        if (Cur_Treble > MAX_TREBLE) Cur_Treble = MAX_TREBLE;
        AWE->Equalizer(Cur_Bass, Cur_Treble);
        Update_treble();
        break;
      case 't': // lower treble
        Cur_Treble -= 1;
        if (Cur_Treble < MIN_TREBLE) Cur_Treble = MIN_TREBLE;
        AWE->Equalizer(Cur_Bass, Cur_Treble);
        Update_treble();
        break;

      case 'B': // raise bass
        Cur_Bass += 1;
        if (Cur_Bass > MAX_BASS) Cur_Bass = MAX_BASS;
        AWE->Equalizer(Cur_Bass, Cur_Treble);
        Update_bass();
        break;
      case 'b': // lower bass
        Cur_Bass -= 1;
        if (Cur_Bass < MIN_BASS) Cur_Bass = MIN_BASS;
        AWE->Equalizer(Cur_Bass, Cur_Treble);
        Update_bass();
        break;

      case 'i': // initial values
        Cur_Volume = MAIN_VOLUME;
        Cur_Bass   = BASS_LEVEL;
        Cur_Treble = TREBLE_LEVEL;
        AWE->MainVolume(Cur_Volume);
        AWE->Equalizer(Cur_Bass, Cur_Treble);
        Update_volume();
        Update_treble();
        Update_bass();
        break;

      case 'q': // Quit
        exit(0); // child exits, parent is signalled with SIGCHLD

      default:
        break;
    } // switch
  } // while
} // Interface::Interactive_control

////////////////////////////////////////////////////////////////////////////////
// Interface::Close_Interface
////////////////////////////////////////////////////////////////////////////////
void Interface::Close_Interface() {
  switch (interfacetype) {
#ifdef INT_NCURSES
    case NCurses:
      endwin();
      break;
#endif
    case Normal:
      break;
  } // switch
} // Interface::Close_Interface

////////////////////////////////////////////////////////////////////////////////
// Interface::Print_header
////////////////////////////////////////////////////////////////////////////////
void Interface::Print_header() {
  switch (interfacetype) {
#ifdef INT_NCURSES
    case NCurses:
      attron(A_BOLD);
      mvaddstr(0, 5, "AWEWavePlayer, plays mixed mode stereo wave data through your EMU8000");
      mvaddstr(1, 5, "Version " PrgVersion " " PrgDate ". Written by Jeroen Witteman");
      mvaddstr(2, 5, "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!");
      attrset(A_NORMAL);
      refresh();
      break;
#endif
    case Normal:
      cout << "AWEWavePlayer, plays mixed mode stereo wave data through your EMU8000" << endl
        << "Version " << PrgVersion << " " << PrgDate << ". Written by Jeroen Witteman" << endl
        << "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!" << endl
        << endl;
      break;
  } // switch
} // Interface::Print_header

////////////////////////////////////////////////////////////////////////////////
// Interface::Print_settings_box
////////////////////////////////////////////////////////////////////////////////
#ifdef INT_NCURSES
void Interface::Print_settings_box() {
  attron(COLOR_PAIR(5));
  mvaddstr(4, 0, "+--------------------------------- SETTINGS -----------------------------------+");
  mvaddstr(5, 0, "|"); mvaddstr(5, 79, "|");
  mvaddstr(6, 0, "|"); mvaddstr(6, 79, "|");
  mvaddstr(7, 0, "|"); mvaddstr(7, 79, "|");
  mvaddstr(8, 0, "|"); mvaddstr(8, 79, "|");
  mvaddstr(9, 0, "+------------------------------------------------------------------------------+");
  attron(COLOR_PAIR(8));
} // Interface::Print_settings_box
#endif

////////////////////////////////////////////////////////////////////////////////
// Interface::Print_settings_text
////////////////////////////////////////////////////////////////////////////////
#ifdef INT_NCURSES
void Interface::Print_settings_text() {
  attron(COLOR_PAIR(8));
  mvaddstr(5, 1, "  SampleSize: ........ bytes (= ..... kB), .... tics");
  mvaddstr(6, 1, " NrOfSamples: ... (Required Memory: ..... kB)");
  mvaddstr(7, 1, "      Chorus: ...        Reverb: ...      Bass: ..        Volume: ...");
  mvaddstr(8, 1, " Chorus Mode: ...   Reverb Mode: ...    Treble: ..   Attenuation: ...");
  attron(COLOR_PAIR(8));
} // Interface::Print_settings_text
#endif

////////////////////////////////////////////////////////////////////////////////
// Print_settings_numbers
////////////////////////////////////////////////////////////////////////////////
#ifdef INT_NCURSES
void Interface::Print_settings_numbers() {
  attron(COLOR_PAIR(7));
  mvprintw(5,15, "%8i", StereoSampleSize);
  mvprintw(5,33, "%5i", StereoSampleSize/1024);
  mvprintw(5,44, "%4i", Time_Stereo_Sample(StereoSampleSize));

  mvprintw(6,15, "%3i", NrOfSamples);
  mvprintw(6,37, "%5i", (StereoSampleSize * NrOfSamples / 1024) + 2);

  mvprintw(7,15, "%3i", settings.chorus);
  mvprintw(7,34, "%3i", settings.reverb);
  mvprintw(7,49, "%2i", settings.bass);
  mvprintw(7,67, "%3i", settings.volume);

  mvprintw(8,15, "%3i", settings.chorus_mode);
  mvprintw(8,34, "%3i", settings.reverb_mode);
  mvprintw(8,49, "%2i", settings.treble);
  mvprintw(8,67, "%3i", settings.attenuation);
  attron(COLOR_PAIR(8));
} // Print_settings_numbers
#endif

////////////////////////////////////////////////////////////////////////////////
// Interface::Print_Available_Memory
// Prints the available memory on your soundcard.
////////////////////////////////////////////////////////////////////////////////
void Interface::Print_available_memory() {
  switch (interfacetype) {
#ifdef INT_NCURSES
    case NCurses:
      mvprintw(6,50, "Available memory: ..... kB");
      attron(COLOR_PAIR(7));
      mvprintw(6,68, "%5i", AWE->AvailableMem()/1024);
      attron(COLOR_PAIR(8));
      refresh();
      break;
#endif
    case Normal:
      cout << "AWE DRAM: " << AWE->AvailableMem()/1024 << " kB memory left." << endl;
      break;
  } // switch

} // Interface::Print_Available_Memory

////////////////////////////////////////////////////////////////////////////////
// Interface::Update_available_memory
////////////////////////////////////////////////////////////////////////////////
void Interface::Update_available_memory() {
  switch (interfacetype) {
#ifdef INT_NCURSES
    case NCurses:
      attron(COLOR_PAIR(7));
      mvprintw(6,68, "%5i", AWE->AvailableMem()/1024);
      attron(COLOR_PAIR(8));
      refresh();
      break;
#endif
    case Normal:
      break;
  } // switch
} // Interface::Update_available_memory

////////////////////////////////////////////////////////////////////////////////
// Interface::Update_volume
////////////////////////////////////////////////////////////////////////////////
void Interface::Update_volume() {
  switch (interfacetype) {
    case NCurses:
      attron(COLOR_PAIR(7));
      mvprintw(7, 67, "%3i", settings.volume);
      attron(COLOR_PAIR(8));
      refresh();
      break;
    case Normal:
      break;
  } // switch
} // Interface::Update_volume

////////////////////////////////////////////////////////////////////////////////
// Interface::Update_treble
////////////////////////////////////////////////////////////////////////////////
void Interface::Update_treble() {
  switch (interfacetype) {
    case NCurses:
      attron(COLOR_PAIR(7));
      mvprintw(8,49, "%2i", settings.treble);
      attron(COLOR_PAIR(8));
      refresh();
      break;
    case Normal:
      break;
  } // switch
} // Interface::Update_treble

////////////////////////////////////////////////////////////////////////////////
// Interface::Update_bass
////////////////////////////////////////////////////////////////////////////////
void Interface::Update_bass() {
  switch (interfacetype) {
    case NCurses:
      attron(COLOR_PAIR(7));
      mvprintw(7,49, "%2i", settings.bass);
      attron(COLOR_PAIR(8));
      refresh();
      break;
    case Normal:
      break;
  } // switch
} // Interface::Update_bass

////////////////////////////////////////////////////////////////////////////////
// Interface::Print_settings
////////////////////////////////////////////////////////////////////////////////
void Interface::Print_settings() {
  switch (interfacetype) {
#ifdef INT_NCURSES
    case NCurses:
      Print_settings_text();
      Print_settings_box();
      Print_settings_numbers();
      refresh();
      break;
#endif
    case Normal:
      cout
        << "---------------------------------- SETTINGS: ----------------------------------" << endl
        << " SampleSize: " << StereoSampleSize << " bytes (= "
        << StereoSampleSize / 1024 << " kB), "
        << Time_Stereo_Sample(StereoSampleSize) << " tics" << endl
        << "NrOfSamples: " << NrOfSamples << " (Required Memory: "
        << (StereoSampleSize * NrOfSamples / 1024) + 2 << " kB)" << endl
        << "     Chorus: " << setw(3) << settings.chorus
        << "        Reverb: " << setw(3) << settings.reverb
        << "     Bass: " << setw(2) << settings.bass
        << "        Volume: " << setw(3) << settings.volume << endl
        << "Chorus mode: " << setw(3) << settings.chorus_mode
        << "   Reverb mode: " << setw(3) << settings.reverb_mode
        << "   Treble: " << setw(2) << settings.treble
        << "   Attenuation: " << setw(3) << settings.attenuation << endl
        << "-------------------------------------------------------------------------------" << endl
        << endl;
      break;
  } // switch
} // Interface::Print_settings

////////////////////////////////////////////////////////////////////////////////
// Interface::Check_argument
// Checks for the argument 'arg' whether the value is in the range from 'from'
// to 'to'. If not it will print an error message and exit.
////////////////////////////////////////////////////////////////////////////////
void Interface::Check_argument(int arg, int from, int to, char* name) {
  if (! In_Range(arg, from, to)) {
#ifdef INT_NCURSES
    if (interfacetype == NCurses) {
      move(0, 0);
      endwin();
    } // if
#endif

    cerr << "The " << name  << " should be from " << from << " to " << to
      << ". (currently set to " << arg << ")" << endl;
    exit(1);
  } // if
} // Interface::Check_argument

////////////////////////////////////////////////////////////////////////////////
// Interface::Check_all_arguments
////////////////////////////////////////////////////////////////////////////////
void Interface::Check_all_arguments() {
  Check_argument(NrOfSamples,          MIN_NROFSAMPLES, MAX_NROFSAMPLES, "number of samples");
  Check_argument(settings.chorus,      MIN_CHORUS,      MAX_CHORUS,      "chorus argument");
  Check_argument(settings.reverb,      MIN_REVERB,      MAX_REVERB,      "reverb argument");
  Check_argument(settings.chorus_mode, MIN_CHORUSM,     MAX_CHORUSM,     "chorus mode");
  Check_argument(settings.reverb_mode, MIN_REVERBM,     MAX_REVERBM,     "reverb mode");
  Check_argument(settings.bass,        MIN_BASS,        MAX_BASS,        "bass value");
  Check_argument(settings.treble,      MIN_TREBLE,      MAX_TREBLE,      "treble value");
  Check_argument(settings.volume,      MIN_VOLUME,      MAX_VOLUME,      "main volume");
} // Interface::Check_all_arguments

////////////////////////////////////////////////////////////////////////////////
// Interface::Check_enough_memory
////////////////////////////////////////////////////////////////////////////////
void Interface::Check_enough_memory() {
  int available = AWE->AvailableMem();
  int required = (StereoSampleSize * NrOfSamples) + 2048;
  if (available < required) {
#ifdef INT_NCURSES
    if (interfacetype == NCurses) {
      move(0, 0);
      endwin();
    } // if
#endif
    cerr << "AWEWavePlayer: Not enough memory available on soundcard." << endl;
    cerr << "Memory available: " << available << " bytes" << endl;
    cerr << "Memory required:  " << required << " bytes" << endl;
    cerr << "Change the sample size and/or the number of samples or reset samples." << endl;
    exit(1);
  } // if
} // Interface::Check_enough_memory

////////////////////////////////////////////////////////////////////////////////
// Interface::Clear_work
////////////////////////////////////////////////////////////////////////////////
#ifdef INT_NCURSES
void Interface::Clear_work(int vos, int i) {
  attron(COLOR_PAIR(8));
  for (int j=0; j < 3; j++) {
    mvaddstr(vos+1+j, 3*i, "..");
  } // for
} // Interface::Clear_work
#endif

////////////////////////////////////////////////////////////////////////////////
// Interface::Show_work
////////////////////////////////////////////////////////////////////////////////
void Interface::Show_work(int i, Work work) {
#ifdef INT_NCURSES
  const int vos = 14; // the vertical offset
  int j;  // counter
  static bool first = true; // is this the first function call?
#endif

  if (verbosity == verbose) {
#ifdef INT_NCURSES
    if (interfacetype == NCurses) {
      if (first) {
        for (j=0; j < NrOfSamples; j++) {
          Clear_work(vos, j);
          mvprintw(vos, 3*j, "%i", j);
        } // for
        first = false;
      } // if
    } // if
#endif

    switch(work) {
      case Reading:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(4));
            mvaddstr(vos+1, 3*i, "rd");
            break;
#endif
          case Normal:
            cout << "[" << i << flush;
            break;
        } // switch
        break;

      case Waiting:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(3));
            mvaddstr(vos+2, 3*i, "WW");
            break;
#endif
          case Normal:
            cout << "W" << flush;
            break;
        } // switch
        break;

      case AllocatingL:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(7));
            mvaddstr(vos+3, 3*i, "A");
            break;
#endif
          case Normal:
            cout << "[" << i << "Al" << flush;
            break;
        } // switch
        break;

      case AllocatingR:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(7));
            mvaddstr(vos+3, 3*i+1, "A");
            break;
#endif
          case Normal:
            cout << "[" << i << "Ar" << flush;
            break;
        } // switch
        break;

      case LoadingL:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(6));
            mvaddstr(vos+3, 3*i, "L");
            break;
#endif
          case Normal:
            cout << "Ll" << flush;
            break;
        } // switch
        break;

      case LoadingR:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(6));
            mvaddstr(vos+3, 3*i+1, "L");
            break;
#endif
          case Normal:
            cout << "[" << i << "Lr" << flush;
            break;
        } // switch
        break;

      case ReplacingL:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(2));
            mvaddstr(vos+3, 3*i, "R");
            break;
#endif
          case Normal:
            cout << "R" << flush;
            break;
        } // switch
        break;

      case ReplacingR:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            attron(COLOR_PAIR(2));
            mvaddstr(vos+3, 3*i+1, "R");
            break;
#endif
          case Normal:
            cout << "R" << flush;
            break;
        } // switch
        break;

      case Done:
        switch (interfacetype) {
#ifdef INT_NCURSES
          case NCurses:
            Clear_work(vos, i);
            break;
#endif
          case Normal:
            cout << "] " << flush;
            break;
        } // switch
        break;
    } // switch
#ifdef INT_NCURSES
    if (interfacetype == NCurses)
      refresh();
#endif
  } // if
} // Interface::Show_work

////////////////////////////////////////////////////////////////////////////////
// Interface::Print
////////////////////////////////////////////////////////////////////////////////
void Interface::Print(Printwhat what) {
  switch (what) {
    case LoadFirst:
      switch (interfacetype) {
#ifdef INT_NCURSES
        case NCurses:
          mvaddstr(12, 0, "Loading first sample and allocating continuous AWE memory blocks."); refresh();
          break;
#endif
        case Normal:
          if (verbosity != quiet) cout << "Loading first sample and allocating continuous AWE memory blocks." << endl;
          break;
      } // switch
      break;

    case Playing:
      switch (interfacetype) {
#ifdef INT_NCURSES
        case NCurses:
          attron(COLOR_PAIR(5)); attron(A_BOLD);
          mvaddstr(11, 0, "Playing");
          attroff(A_BOLD); attron(COLOR_PAIR(8)); refresh();
          break;
#endif
        case Normal:
          if (verbosity != quiet) cout << endl << "Playing started." << endl;
          break;
      } // switch
      break;

    case LoadRest:
      switch (interfacetype) {
#ifdef INT_NCURSES
        case NCurses:
          move(12,0); clrtoeol();
          mvaddstr(12, 0, "Loading rest of samples"); refresh();
          break;
#endif
        case Normal:
          if (verbosity != quiet) cout << "Loading rest of samples" << (verbose ? ": " : "... ") << flush;
          break;
      } // switch
      break;

    case Replacing:
      switch (interfacetype) {
#ifdef INT_NCURSES
        case NCurses:
          move(12,0); clrtoeol();
          mvaddstr(12, 0, "Replacing samples"); refresh();
          break;
#endif
        case Normal:
          if (verbosity == verbose) cout << endl << "Replacing samples: " << flush;
          break;
      } // switch
      break;

    case DoneReplacing:
      switch (interfacetype) {
#ifdef INT_NCURSES
        case NCurses:
          move(12,0); clrtoeol();
          mvaddstr(12, 0, "Done replacing samples, waiting for playing to finish..."); refresh();
          break;
#endif
        case Normal:
          if (verbose) cout << endl;
          if (verbosity != quiet) cout << "Done replacing samples, waiting for playing to finish..." << endl;
          break;
      } // switch
      break;

    case Unloading:
      switch (interfacetype) {
#ifdef INT_NCURSES
        case NCurses:
          move(12,0); clrtoeol();
          attron(COLOR_PAIR(8));
          mvaddstr(12, 0, "Unloading samples"); refresh();
          break;
#endif
        case Normal:
          if (verbosity != quiet) cout << "Unloading samples... " << flush;
          break;
      } // switch
      break;

    case Finished:
      switch (interfacetype) {
#ifdef INT_NCURSES
        case NCurses:
          break;
#endif
        case Normal:
          if (verbosity != quiet) cout << endl << "** AWEWavePlayer done ***" << endl;
          break;
      } // switch
      break;
  } // switch
} // Interface::Print

////////////////////////////////////////////////////////////////////////////////
// Interface::Usage
// Prints the usage on the screen
////////////////////////////////////////////////////////////////////////////////
void Interface::Usage(char* command) {
  cerr
    << "Usage examples: " << endl
    << "(1) mpg123 -s -q <filename.mp3> | " << command << " [options]" << endl
    << "(2) " << command << " [options] < <sample.wav>" << endl
    << endl
    << command << " options:" << endl
    << "-h | --help                Display this help information       " << endl
    << "-i | --initialize          Reset all samples                   " << endl
    << "-v | --verbose             Verbose mode                        " << endl
#ifdef INT_NCURSES
    << "-nc| --ncurses             Ncurses interface                   " << endl
#endif
    << "-q | --quiet               Quiet mode                          " << endl
    << "-B | --sample-bytes <size> Size of samples in bytes (default "   << SAMPLESIZE   << ")" << endl
    << "-S | --sample-secs <secs>  Size of samples in seconds          " << endl
    << "-n | --#samples <nr>       Number of samples [2-64] (default "   << NROFSAMPLES  << ")" << endl
    << "-V | --volume <vol>        Set the main volume [0-127] (default "<< MAIN_VOLUME  << ")" << endl
    << "-a | --attenuation <val>   Set the initial attenuation (default "<< ATTENUATION  << ")" << endl
    << "-b | --bass <nr>           Set bass level [0-11] (default "      << BASS_LEVEL   << ")" << endl
    << "-t | --treble <nr>         Set treble level [0-11] (default "    << TREBLE_LEVEL << ")" << endl
    << "-c | --chorus <nr>         Amount of chorus [0-127] (default "   << CHORUS       << ")" << endl
    << "-r | --reverb <nr>         Amount of reverb [0-127] (default "   << REVERB       << ")" << endl
    << "-C | --chorus-mode <mode>  Set the chorus mode [0-7] (default "  << CHORUS_MODE  << ")" << endl
    << "-R | --reverb-mode <mode>  Set the reverb mode [0-7] (default "  << REVERB_MODE  << ")" << endl
    << "-d | --debug <mode>        Set the debug mode [0-9] (default 0)" << endl
    << endl
    << "Possible chorus modes:         Possible reverb modes:     " << endl
    << "0 = chorus1  4 = feedback      0 = room1  4 = hall2       " << endl
    << "1 = chorus2  5 = flanger       1 = room2  5 = plate       " << endl
    << "2 = chorus3  6 = shortdelay    2 = room3  6 = delay       " << endl
    << "3 = chorus4  7 = shortdelay2   3 = hall1  7 = panningdelay" << endl
    << endl;

  exit(0);
} // Interface::Usage

///////////////////////////////////// EOF //////////////////////////////////////
