////////////////////////////////////////////////////////////////////////////////
// player.cc - AWEWavePlayer version 0.3 by Jeroen Witteman
////////////////////////////////////////////////////////////////////////////////
//
// This program can be used to play samples through the EMU8000 of your Sound
// Blaster AWE32 (and family) soundcard. It reads 16 bit 44 kHz mixed mode
// stereo wave data from standard input. The sample will be split into pieces
// of 'SampleSize' bytes and loaded into the memory of your AWE. After loading
// 'NrOfSamples' of these samples playing will be started. During playing the
// sample that has just been played will be replace with a new sample read from
// standard input.
//
// Written by:    Jeroen Witteman
// Email adress:  jwittema@wi.leidenuniv.nl
// WWW Homepage:  http://www.wi.leidenuniv.nl/home/jwittema/
//
// copyright (c) 1998 by Jeroen Witteman, all rights reserved.
// THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!
////////////////////////////////////////////////////////////////////////////////

// includes
#include <iostream.h>  // used for cin and cout
#include <stdlib.h>    // used for atoi() and perror()
#include <unistd.h>    // used for usleep()
#include <getopt.h>    // used for getopt() (of course !)
#include <signal.h>
#ifdef INT_NCURSES
#include <ncurses.h>
#endif
#include <sys/times.h> // used for times()
#include "awe_sequencer.h"
#include "interface.h"
#include "common.h"
#include "seq.h"
#include "version.h"
#include "config.h"

// global variables
Interface* UI;                      // pointer to the User Interface
AWE_Sequencer* AWE;
int StereoSampleSize = SAMPLESIZE;  // sample size in bytes
int NrOfSamples = NROFSAMPLES;      // number of samples
Settings settings;                  // settings

////////////////////////////////////////////////////////////////////////////////
// Read_Data
// Reads mixed stereo sample data (interleaved stereo sample) from standard
// input and splits this into two samples: one for the left channel and one for
// the right channel. The pointers Left and Right should point to an array of
// MonoSampleSize bytes long.
// Returns how may bytes were really read from standard input.
////////////////////////////////////////////////////////////////////////////////
int Read_Data(char* Left, char* Right) {
  register int i; // counter
  register int j; // counter
  int NrOfBytesRead = 0; // how many bytes did we read from standard input
  char* tmp;

  tmp = new char[StereoSampleSize]; // allocate temporary memory

  cin.read(tmp, StereoSampleSize); // read StereoSampleSize bytes from standard input to tmp
  NrOfBytesRead = cin.gcount(); // the number of bytes we just read
  if (cin.eof()) {  // if we're at the end of standard input
    for (i=StereoSampleSize-1; i>NrOfBytesRead; i--) {
      tmp[i]=0;  // remove old sample data (or garbage) from memory
    } // for
  } // if

  // split the mixed mode sample data into two samples
  for (i=0; i<StereoSampleSize/2; i+=2) {
    j=2*i;
    Left[i]    = tmp[j];
    Left[i+1]  = tmp[j+1];
    Right[i]   = tmp[j+2];
    Right[i+1] = tmp[j+3];
  } // for

  delete[] tmp; // deallocate temporary memory

  return NrOfBytesRead;
} // Read_Data

////////////////////////////////////////////////////////////////////////////////
// Load_initial_samples
// Allocates continuous samples and loads the initial sample data, then starts
// playing.
// Returns the time when the first sample can be replace
// (= Time when playing started + sample time).
////////////////////////////////////////////////////////////////////////////////
clock_t Load_initial_samples(char* LeftWaveData, char* RightWaveData) {
  const int Sample_timetics = Time_Stereo_Sample(StereoSampleSize);
  const int Header_Size = AWE_PATCH_INFO_SIZE + AWE_SAMPLE_INFO_SIZE; // 48
  int readbytes = 0;
  clock_t Replace_time = 0;
  int i = 0; // counter

  UI->Print(LoadFirst);
  UI->Show_work(i, Reading);
  readbytes = Read_Data(LeftWaveData + Header_Size, RightWaveData + Header_Size);

  if (readbytes > 0) {
    // first allocate the memory blocks: load (empty) samples for left samples, than for right samples.
    for (i = 0; i < NrOfSamples; i++) {
      UI->Show_work(i, (i==0) ? LoadingL : AllocatingL);
      AWE->LoadSample(LeftWaveData, i, NOREPLACE); 
      UI->Update_available_memory();
      UI->Show_work(i, Done);
    } // for
    for (i = 0; i < NrOfSamples; i++) {
      UI->Show_work(i, (i==0) ? LoadingR : AllocatingR);
      AWE->LoadSample(RightWaveData, i+IDOS, NOREPLACE); 
      UI->Update_available_memory();
      UI->Show_work(i, Done);
    } // for

    // AWE->LoadVoiceInfo(sample_id, banknr, instr#,     pan)
    AWE->LoadVoiceInfo(           0,      0,      0, PANLEFT);
    AWE->LoadVoiceInfo(        IDOS,      0,      0,PANRIGHT);
    UI->Update_available_memory();

    AWE->StartPlaying();
    // time when we can replace the first sample = current time + sample time
    Replace_time = times(NULL) + Sample_timetics; 
    UI->Print(Playing);

    // Initial loading of samples (sample 0 already loaded)
    UI->Print(LoadRest);
    for (i=1; i<NrOfSamples; i++) {
      if (readbytes > 0) {
        UI->Show_work(i, Reading);
        readbytes=Read_Data(LeftWaveData + Header_Size, RightWaveData + Header_Size);
        UI->Show_work(i, ReplacingL);
        AWE->LoadSample( LeftWaveData,    i, REPLACE); 
        UI->Show_work(i, ReplacingR);
        AWE->LoadSample(RightWaveData, i+IDOS, REPLACE);
        UI->Show_work(i, Done);
      } // if
    } // for

    UI->Print_available_memory();

    return Replace_time;
  } // if
  return 0;
} // Load_initial_samples

////////////////////////////////////////////////////////////////////////////////
// Start_Playing
// Starts the playing process. First allocating continuous memory blocks in the
// AWE's memory (see below). This is neccesary to make an instrument which
// loops over all samples. After the memory allocation the real sample data
// will be loaded into the AWE's memory. Then a note is started playing, this
// will continue on looping over all the samples. When it has reached the end
// of the first sample, the first sample will be replaced by new sample data
// read from standard input.
//
// |--L1--||--L2--| ... |--Ln--||--R1--||--R2--| ... |--Rn--|   AWE memory
//
// \______ sample loop ________/\______ sample loop ________/
//              \                           /
//               \_____ one instrument ____/
// 
// L1 = first left sample,   R1 = first right sample
// L2 = second left sample,  R2 = second right sample
// etc.
////////////////////////////////////////////////////////////////////////////////
void Start_Playing() {
  const int Sample_timetics = Time_Stereo_Sample(StereoSampleSize);
  const int Header_Size = AWE_PATCH_INFO_SIZE + AWE_SAMPLE_INFO_SIZE; // 48
  char* LeftWaveData;      // pointer to the left channel wave data
  char* RightWaveData;     // pointer to the right channel wave data
  clock_t Replace_time=0;  // for storing the time when we can replace a sample
  int readbytes = 0;       // how many bytes did we read from standard input
  int lastsamplebytes = 0; // how many bytes was the last sample 
  int x = 0;               // temporary storage
  int i =0;                // counter

  // allocate memory
  LeftWaveData  = new char[MonoSampleSize + Header_Size];
  RightWaveData = new char[MonoSampleSize + Header_Size];

  AWE->OpenPatch();

  Replace_time = Load_initial_samples(LeftWaveData, RightWaveData);
  if (Replace_time > 0) { 
    UI->Update_available_memory();

    UI->Print(Replacing);
    // already read sample data from standard input
    UI->Show_work(i, Reading);
    readbytes=Read_Data(LeftWaveData + Header_Size, RightWaveData + Header_Size);
    UI->Show_work(i, Waiting);
    // and wait until first sample has finished
    usleep(( (x=(Replace_time - times(NULL)) * 10000) ) > 0 ? x : 0);

    // Replacing process
    i = 0;
    while (readbytes > 0) {
      Replace_time += Sample_timetics; // time when we can replace the next sample
      UI->Show_work(i, ReplacingL);
      AWE->LoadSample( LeftWaveData,    i, REPLACE);
      UI->Show_work(i, ReplacingR);
      AWE->LoadSample(RightWaveData, i+IDOS, REPLACE);
      UI->Show_work(i, Done);

      UI->Show_work((i+1)%NrOfSamples, Reading);
      readbytes=Read_Data(LeftWaveData + Header_Size, RightWaveData + Header_Size);
      // store size of last sample, so that we know when to stop playing
      if(readbytes < StereoSampleSize && readbytes > 0) lastsamplebytes = readbytes;
      UI->Show_work((i+1)%NrOfSamples, Waiting);
      usleep(( (x=(Replace_time - times(NULL)) * 10000) ) > 0 ? x : 0);
      if (++i == NrOfSamples) i = 0;
    } // while
    UI->Print(DoneReplacing);
  } // if

  // Almost done
  delete[] LeftWaveData;    // deallocate memory
  delete[] RightWaveData;   // 

  // wait until all samples are finished 
  Replace_time += ((NrOfSamples-2) * Sample_timetics) + Time_Stereo_Sample(lastsamplebytes);
  usleep(( (x=(Replace_time - times(NULL)) * 10000) ) > 0 ? x : 0);

  AWE->ClosePatch();

  AWE->ReleaseAllNotes();
  UI->Print(Unloading);
  for (i=0; i<NrOfSamples; i++) {
    AWE->UnloadSample();
    UI->Update_available_memory(); 
  } // for
} // Start_Playing

////////////////////////////////////////////////////////////////////////////////
// Signal_Handler
// This function handles the SIGCHLD signal.
////////////////////////////////////////////////////////////////////////////////
void Signal_Handler(int Signal) {
  switch (Signal) {
    case SIGCHLD:  // If the child (= UI) exited, exit too
      AWE->Shutdown();
      UI->Close_Interface();
      UI->Print(Finished);
      exit(0);
  } // switch
} // Signal_Handler

////////////////////////////////////////////////////////////////////////////////
// main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv) {
  UI = new Interface;
  AWE = new AWE_Sequencer;
  int i = 1;                      // counter
  const option Options[] = {      // the long option list
    {"help",         no_argument,       0, 'h'},
    {"initialize",   no_argument,       0, 'i'},
    {"verbose",      no_argument,       0, 'v'},
    {"quiet",        no_argument,       0, 'q'},
    {"sample-bytes", required_argument, 0, 'B'},
    {"sample-secs",  required_argument, 0, 'S'},
    {"#samples",     required_argument, 0, 'n'},
    {"bass",         required_argument, 0, 'b'},
    {"treble",       required_argument, 0, 't'},
    {"chorus",       required_argument, 0, 'c'},
    {"reverb",       required_argument, 0, 'r'},
    {"chorus-mode",  required_argument, 0, 'C'},
    {"reverb-mode",  required_argument, 0, 'R'},
    {"debug",        required_argument, 0, 'd'},
    {"volume",       required_argument, 0, 'V'},
    {"attenuation",  required_argument, 0, 'a'},
    {"nc",           no_argument,       0, 'X'},
    {"ncurses",      no_argument,       0, 'X'},
    {0,0,0,0} // last element has to be filled with zeroes
  };
  
  // parse options
  while ((i=getopt_long_only(argc, argv, "hivqn:B:S:n:b:t:c:r:C:R:d:V:a:", Options, NULL)) > 0) {
    switch(i) {
      case '?':
      case 'h': UI->Usage(argv[0]);                  break; // print usage and exit
      case 'i': settings.reset       = true;         break; // reset samples
      case 'v': UI->verbosity        = verbose;      break; // set verbose on
      case 'q': UI->verbosity        = quiet;        break; // set quiet on
      case 'B': StereoSampleSize = atoi(optarg);     break; // set sample size in bytes
      case 'S': StereoSampleSize = atoi(optarg) * 4 * 44100 ; break; // set sample size in seconds
      case 'n': NrOfSamples          = atoi(optarg); break; // set # of samples
      case 'b': settings.bass        = atoi(optarg); break; // set bass
      case 't': settings.treble      = atoi(optarg); break; // set treble
      case 'c': settings.chorus      = atoi(optarg); break; // set chorus
      case 'r': settings.reverb      = atoi(optarg); break; // set reverb
      case 'C': settings.chorus_mode = atoi(optarg); break; // set chorus mode
      case 'R': settings.reverb_mode = atoi(optarg); break; // set reverb mode
      case 'd': settings.debugmode   = atoi(optarg); break; // set debugmode
      case 'V': settings.volume      = atoi(optarg); break; // set main volume
      case 'a': settings.attenuation = atoi(optarg); break; // set initial attenuation
#ifdef INT_NCURSES
      case 'X': UI->interfacetype    = NCurses;      break; // set the interface to ncurses
#endif
      default: UI->Usage(argv[0]);                   break; // print usage and exit
    } // switch
  } // while

  UI->Initialize_interface();
  UI->Print_header();
  UI->Print_settings();

  UI->Check_all_arguments();

  AWE->Initialize();
  UI->Print_available_memory();
  AWE->ResetSamples(); // reset samples (if chosen)
  UI->Update_available_memory();
  UI->Check_enough_memory();

  if (settings.debugmode >= 0) 
    AWE->DebugMode(settings.debugmode); // set the debug mode
  AWE->Equalizer(settings.bass, settings.treble); // set bass and treble
  AWE->ReverbMode(settings.reverb_mode); // set reverb mode
  AWE->ChorusMode(settings.chorus_mode); // set chorus mode
  AWE->InitialAttenuation(settings.attenuation); // set initial attenuation

  if (UI->interfacetype == NCurses) {
    pid_t pid;
    switch(pid = fork()) {
      case -1: // error
        perror("Error during fork()");
        exit(1);
      case 0: // child
        UI->Interactive_control();
        exit(0);
      default: // parent
        signal(SIGCHLD, Signal_Handler); // install the signal handler for SIGCHLD
        Start_Playing();
        if (kill(pid, SIGKILL) == -1) { // kill the child process
          perror("Error killing child");
        } // if
    } // switch
  } // if
  else {
    Start_Playing();
  } // else

  AWE->Shutdown();
  UI->Close_Interface();
  UI->Print(Finished);

  return 0;
} // main

///////////////////////////////////// EOF //////////////////////////////////////
